/**************************************************************************
 *
 * invf_get.c -- Inverted file reading routines
 * Copyright (C) 1994  Neil Sharman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: invf_get.cpp 655 1999-10-11 02:58:42Z cs025 $
 *
 **************************************************************************/

#include "sysfuncs.h"

#include "memlib.h"
#include "filestats.h"
#include "timing.h"
#include "bitio_gen.h"
#include "bitio_m_abstract.h"
#include "bitio_m_mem.h"
#include "bitio_m.h"
#include "sptree.h"
#include "messages.h"
#include "netorder.h"  /* [RPAP - Jan 97: Endian Ordering] */

#include "mg.h"
#include "invf.h"
#include "text.h"
#include "lists.h"
#include "backend.h"
#include "invf_get.h"
#include "mg_errors.h"
#include "DocEntry.h"

#define SKIPPING
#define SHORT_CUT 1

struct pool
  {
    int length, pos;
    struct pool *next;
    Invf_Doc_Entry ide[1];
  };

static int skip_header = 0;

static char skip_header_text0[] =
"#######################################################################\n"
"#\n"
"# 1\tWord number.\n"
"# 2\tNumber of time the word occurs in the collection.\n"
"# 3\tNumber of documents the word occurs in.\n"
"# 4\tThe word.\n"
"# 5\tNumber of docnum/word_count entries decoded.\n"
"# 6\tNumber of hits.\n"
"# 7\tNumber of entries in the splay tree or hash table.\n"
"# 8\tAmount added to the accumulators while while processing this word.\n"
"#\n"
"#######################################################################\n";

static char skip_header_text1[] =
"#######################################################################\n"
"#\n"
"# 1\tWord number.\n"
"# 2\tNumber of time the word occurs in the collection.\n"
"# 3\tNumber of documents the word occurs in.\n"
"# 4\tThe word.\n"
"# 5\tNumber of skips taken in evaluating the inverted file entry.\n"
"# 6\tNumber of docnum/word_count entries decoded.\n"
"# 7\tNumber of hits.\n"
"# 8\tNumber of entries in the splay tree or hash table.\n"
"# 9\tAmount added to the accumulators while while processing this word.\n"
"#\n"
"#######################################################################\n";


static char skip_header_text2[] =
"#######################################################################\n"
"#\n"
"#  1\tWord number.\n"
"#  2\tNumber of time the word occurs in the collection.\n"
"#  3\tNumber of documents the word occurs in.\n"
"#  4\tThe word.\n"
"#  5\tSkip size.\n"
"#  6\tNumber of skips taken in evaluating the inverted file entry.\n"
"#  7\tNumber of docnum/word_count entries decoded.\n"
"#  8\tNumber of hits.\n"
"#  9\tNumber of entries in the splay tree or hash table.\n"
"# 10\tAmount added to the accumulators while while processing this word.\n"
"#\n"
"#######################################################################\n";

static char *skip_header_text[] =
{skip_header_text0,
 skip_header_text1,
 skip_header_text2};


#define COSINE_STARTTERM \
if (cs.startTerm() != 0)                                     \
  break;                                                     \
                                                             \
DECODE_START (cs.buffer, cs.we->invf_len);                   \
                                                             \
cs.WordLog = log ((double) cs.num_docs / (double) cs.we->max_doc_count);  \
                                                             \
cs.Wqt = (rqi->QueryFreqs ? Terms->TE[cs.term].Count : 1) * cs.WordLog;


/**************************************************************************
 *
 * Code for the hash table.
 *
 */

static Hash_Table *
HT_create (query_data * qd, unsigned max_entries,
	   unsigned max_size)
{
  unsigned es, hts;
  Hash_Table *HT;
  if (max_entries == 0)
    max_entries = 8192;
  es = sizeof (Hash_Table) + sizeof (Invf_Doc_EntryH) * (max_entries - 1);
  hts = sizeof (Invf_Doc_EntryH *) * max_size;

  HT = (Hash_Table *) Xmalloc (es);
  if (!HT)
    return NULL;

  HT->HashTable = (Invf_Doc_EntryH **) Xmalloc (hts);
  if (!(HT->HashTable))
    {
      Xfree (HT);
      return NULL;
    }

  ChangeMemInUse (qd, es + hts);
  bzero ((char *) (HT->HashTable), hts);
  HT->max = max_entries;
  HT->num = 0;
  HT->Suplimentary_Entries = NULL;
  HT->Suplimentary_Size = 0;
  HT->Suplimentary_Num = 0;
  HT->Head = NULL;
  HT->size = max_size;
  return HT;
}


static void 
HT_Sup_create (query_data * qd, Hash_Table * HT, unsigned max_size)
{
  int es = sizeof (Invf_Doc_EntryH) * max_size;
  HT->Suplimentary_Entries = (Invf_Doc_EntryH *) Xmalloc (es);
  if (!HT->Suplimentary_Entries)
    return;
  HT->Suplimentary_Size = max_size;
  HT->Suplimentary_Num = 0;
  ChangeMemInUse (qd, es);
}

static void 
HT_Sup_Add (Hash_Table * HT, unsigned long DN, float Sum)
{
  Invf_Doc_EntryH *IDEH;
  IDEH = &HT->Suplimentary_Entries[HT->Suplimentary_Num++];
  IDEH->next = NULL;
  IDEH->DocNum = DN;
  IDEH->Weight = Sum;
}

abstractDocEntry *
HT_find (Hash_Table * HT, unsigned long DN)
{
  Invf_Doc_EntryH *IDEH;
  if (!HT->HashTable)
    FatalError (1, "Accessing the hash_table through HT_find after HT_sort");
  IDEH = HT->HashTable[DN % HT->size];
  while (IDEH)
    {
      if (IDEH->DocNum == DN)
	return IDEH;
      IDEH = IDEH->next;
    }
  return NULL;
}


static int 
HT_insert (Hash_Table * HT,
	   unsigned long DN, float Sum)
{
  Invf_Doc_EntryH **IDEH;
  if (HT->num == HT->max)
    return 0;
  IDEH = &(HT->HashTable[DN % HT->size]);
  while (*IDEH)
    IDEH = &((*IDEH)->next);
  *IDEH = &(HT->IDEH[HT->num++]);
  (*IDEH)->next = NULL;
  (*IDEH)->DocNum = DN;
  (*IDEH)->Weight = Sum;
  return 1;
}


void 
HT_free (query_data * qd, Hash_Table * HT)
{
  unsigned es;
  es = sizeof (Hash_Table) + sizeof (Invf_Doc_EntryH) * HT->max +
    sizeof (Invf_Doc_EntryH *) * HT->size;
  if (HT->HashTable)
    Xfree (HT->HashTable);
  if (HT->Suplimentary_Entries)
    {
      Xfree (HT->Suplimentary_Entries);
      es += sizeof (Invf_Doc_EntryH) * HT->Suplimentary_Size;
    }
  Xfree (HT);
  ChangeMemInUse (qd, -es);
}


static int 
IDEH_comp (const void *a, const void *b)
{
  const Invf_Doc_EntryH *A = (Invf_Doc_EntryH *) a;
  const Invf_Doc_EntryH *B = (Invf_Doc_EntryH *) b;
  return A->DocNum - B->DocNum;
}

static void 
HT_sort (query_data * qd, Hash_Table * HT)
{
  int i;
  Invf_Doc_EntryH *l, *r, **b;
  unsigned es = sizeof (Invf_Doc_EntryH *) * HT->size;
  if (HT->HashTable)
    Xfree (HT->HashTable);
  HT->size = 0;
  HT->HashTable = NULL;
  ChangeMemInUse (qd, -es);

  qsort (HT->IDEH, HT->num, sizeof (Invf_Doc_EntryH), IDEH_comp);

  for (i = 0; i < HT->num - 1; i++)
    HT->IDEH[i].next = &HT->IDEH[i + 1];
  HT->IDEH[i].next = NULL;
  l = HT->IDEH;

  if (HT->Suplimentary_Entries)
    {
      for (i = 0; i < HT->Suplimentary_Num - 1; i++)
	HT->Suplimentary_Entries[i].next = &HT->Suplimentary_Entries[i + 1];
      HT->Suplimentary_Entries[i].next = NULL;
    }
  r = HT->Suplimentary_Entries;

  b = &HT->Head;

  while (l || r)
    {
      if (r && (!l || (l->DocNum > r->DocNum)))
	{
	  *b = r;
	  b = &r->next;
	  r = *b;
	}
      else
	{
	  *b = l;
	  b = &l->next;
	  l = *b;
	}
    }
  *b = NULL;
}






/**************************************************************************
 *
 * Code for the hash table.
 *
 */

static List_Table *
LT_create (query_data * qd, unsigned max_entries)
{
  unsigned es;
  List_Table *LT;
  if (max_entries == 0)
    max_entries = 8192;
  es = sizeof (List_Table) + sizeof (Invf_Doc_Entry) * (max_entries - 1);

  LT = (List_Table *) Xmalloc (es);
  if (!LT)
    return NULL;

  ChangeMemInUse (qd, es);
  LT->max = max_entries;
  LT->num = 0;
  return LT;
}



Invf_Doc_Entry *
LT_find (List_Table * LT, unsigned long DN)
{
  int l = 0;
  int r = LT->num - 1;
  int m, c;
  while (l <= r)
    {
      m = (l + r) / 2;
      c = DN - LT->IDE[m].DocNum;
      if (c < 0)
	r = m - 1;
      else if (c > 0)
	l = m + 1;
      else
	return &LT->IDE[m];
    }
  return NULL;
}

static List_Table *
LT_add (query_data * qd, List_Table * LT,
	unsigned long DN, float Sum)
{
  Invf_Doc_Entry *ide;

  /* increase the size of the list if there is not enough */
  /* room for this entry */
  if (LT->num == LT->max)
    {
      int old, enew, max_entries;
      old = sizeof (List_Table) + sizeof (Invf_Doc_Entry) * LT->max;
      max_entries = LT->max + (LT->max >> 1);
      enew = sizeof (List_Table) + sizeof (Invf_Doc_Entry) * max_entries;

      LT = (List_Table *)Xrealloc (LT, enew);
      if (!LT)
	return NULL;

      ChangeMemInUse (qd, enew - old);
      LT->max = max_entries;
    }

  /* add the entry to the end of the list */
  ide = &LT->IDE[LT->num];
  ide->DocNum = DN;
  ide->Weight = Sum;
  LT->num++;

  return LT;
}


void 
LT_free (query_data * qd, List_Table * LT)
{
  unsigned es;
  es = sizeof (List_Table) + sizeof (Invf_Doc_Entry) * LT->max;
  Xfree (LT);
  ChangeMemInUse (qd, -es);
}


static int 
IDE_comp (const void *a, const void *b)
{
  const Invf_Doc_Entry *A = (Invf_Doc_Entry *) a;
  const Invf_Doc_Entry *B = (Invf_Doc_Entry *) b;
  return A->DocNum - B->DocNum;
}

static void 
LT_sort (List_Table * LT)
{
  qsort (LT->IDE, LT->num, sizeof (Invf_Doc_Entry), IDE_comp);
}


/* [RJM 07/97: Ranked Required Terms] */
/* accumulators with a sum less than or equal to 0.0 are to be deleted */
static void 
LT_pack (List_Table * LT)
{
  int s, d = 0;
  for (s = 0; s < LT->num; s++) {
      if (LT->IDE[s].Weight <= 0.0) {
	/* ignore this accumulator */

      } else if (d > 0 && LT->IDE[s].DocNum == LT->IDE[d-1].DocNum) {
	LT->IDE[d-1].Weight += LT->IDE[s].Weight;

      } else {
	if (d != s) LT->IDE[d] = LT->IDE[s];
	d++;
      }
  }

  if (LT->num) LT->num = d;
}


/**************************************************************************/



invf_data *
InitInvfFile (File * InvfFile, stemmed_dict * sd)
{
  invf_data *id;
  if (!(id = (invf_data *) Xmalloc (sizeof (invf_data))))
    {
      mg_errno = MG_NOMEM;
      return NULL;
    }

  id->InvfFile = InvfFile;

  Fread (&id->ifh, sizeof (id->ifh), 1, InvfFile);
  /* [RPAP - Jan 97: Endian Ordering] */
  {
    int i;
    
    NTOHUL(id->ifh.no_of_words);
    NTOHUL(id->ifh.no_of_ptrs);
    NTOHUL(id->ifh.skip_mode);
    for (i = 0; i < 16; i++)
      NTOHUL(id->ifh.params[i]);
    NTOHUL(id->ifh.InvfLevel);
  }

  id->N = sd->sdh.num_of_docs;
  id->Nstatic = sd->sdh.static_num_of_docs;

  return (id);
}


void 
FreeInvfData (invf_data * id)
{
  Xfree (id);
}

/****************************************************************************/


static void 
CalcBlks (query_data * qd, WordEntry * we, int *blk,
	  int *dn_blk, int *len_blk, int *k, int *p)
{
  *p = we->doc_count;

  *blk = BIO_Bblock_Init (qd->id->Nstatic, *p);
  switch (qd->id->ifh.skip_mode)
    {
    case 1:
      {
	unsigned long len;
	if (*p <= qd->id->ifh.params[0])
	  *k = 0;
	else
	  {
	    *k = qd->id->ifh.params[0];
	    len = BIO_Bblock_Bound (qd->id->N, *p);
	    if (qd->id->ifh.InvfLevel >= 2)
	      len += we->count;
	    *dn_blk = BIO_Bblock_Init (qd->sd->sdh.num_of_docs, (*p + *k - 1) / *k);
	    *len_blk = BIO_Bblock_Init (len, (*p + *k - 1) / *k);
	  }
	break;
      }
    case 2:
      {
	unsigned long len;
	if (*p <= qd->id->ifh.params[1])
	  *k = 0;
	else
	  {
	    *k = (int) (2 * sqrt (((double) (*p)) / qd->id->ifh.params[0]));
	    if (*k <= qd->id->ifh.params[1])
	      *k = qd->id->ifh.params[1];
	    len = BIO_Bblock_Bound (qd->id->N, *p);
	    if (qd->id->ifh.InvfLevel >= 2)
	      len += we->count;
	    *dn_blk = BIO_Bblock_Init (qd->sd->sdh.num_of_docs,
				       (*p + *k - 1) / *k);
	    *len_blk = BIO_Bblock_Init (len, (*p + *k - 1) / *k);
	  }
	break;
      }
    default:
      *k = 0;
    }
}


struct cosineSearch
{
  int         term;
  int         num_docs;
  int         num_terms;
  int         MoreAccum;
  int         Anding;
  FILE *      sk;
  u_char      indent;
  query_data *qd;
  TermList *  Terms;
  RankedQueryInfo *rqi;

  // variables for inside term loop;
  int Abort;
  u_char *buffer;
  unsigned long pos; 
  int i;                        /* Pointer into inverted list */
  unsigned long CurrDocNum;	/* NOTE: Document numbers start at 1 */
  double Wqt, WordLog;
  int dn_blk; 
  int len_blk;
  int skip_size, block_counter;
  int blk;			/* bblock control parameter */
  int doc_freq;			/* The number of documents the word occurs in */
  WordEntry *we;
  int next_i, next_doc_num, next_start;
  int skips_taken, ptrs_dec;
  double total_added;
  int hits;

  // variables inside term document loop
  int Count;                      
  double Wdt;
  unsigned long diff;
  abstractDocEntry *mem;

  inline void noteHit(float *local)
  {
    *local = *local + (Wqt * Wdt);
    total_added += Wqt * Wdt;
    hits ++;
  }

  inline void postSkip(bitio_buffer *decodeBuffer)
  {
    if (skip_size != 0 && block_counter == 0 && i != doc_freq - 1)
      {
	CurrDocNum = next_doc_num;
      }
    else
      {
	diff = decodeBuffer->bblock_decode(blk, &pos);
	CurrDocNum += diff;
      }

    if (qd->id->ifh.InvfLevel >= 2)
      Count = decodeBuffer->gamma_decode(&pos);
    else
      Count = (int) ((double) we->count / doc_freq);
    qd->num_of_ptrs++;
    ptrs_dec++;
    i++;
    block_counter--;
    if (block_counter < 0)
      block_counter = skip_size - 1;
  }

  inline int skip(bitio_buffer *decodeBuffer, int nextDoc, int allowShortcut)
  {
    if (skip_size != 0)
      {
	if (Anding && i + skip_size < doc_freq && nextDoc > next_doc_num &&
	    next_doc_num >= 0)
	  {
	  ShortCut:
	    i = next_i;
	    decodeBuffer->seek(next_start);
	    pos = next_start;
	    CurrDocNum = next_doc_num;
	    qd->hops_taken++;
	    block_counter = skip_size - 1;
	    skips_taken ++;
	  }
	if (block_counter == skip_size - 1)
	  {
	    if (i + skip_size < doc_freq)
	      {
		int doc_diff, pos_diff;
		doc_diff = decodeBuffer->bblock_decode(dn_blk, &pos);
		next_doc_num += doc_diff;
		next_i += skip_size;
		pos_diff = decodeBuffer->bblock_decode(len_blk, &pos);
		next_start = pos + pos_diff;
	      }
	    else
	      {
		next_doc_num = -1;
	      }
	  }

	if (Anding && i + skip_size < doc_freq && nextDoc > next_doc_num &&
	    next_doc_num >= 0)
	  {
	    if (allowShortcut)
	      goto ShortCut;
	    else
	      return 1;
	  }
      }
    return 0;
  }

  inline void prepareSkip()
  {
    i             = 0;
    pos           = 0;
    next_i        = 0;
    next_start    = 0;
    next_doc_num  = 0;
    block_counter = skip_size - 1;
  }

  inline int startTerm()
  {
    CalcBlks (qd, we, &blk, &dn_blk, &len_blk, &skip_size, &doc_freq);

    /* -- allocate a buffer big enough to hold the inverted */
    if (!(buffer = (u_char *) Xmalloc (we->invf_len)))
      return 1;
       
    ChangeMemInUse (qd, we->invf_len);
    if (sk && qd->id->ifh.skip_mode == 2)
      fprintf (sk, "%4d", skip_size);

    /* --- Read an inverted file entry for the given word --- */
    Fseek (qd->File_invf, we->invf_ptr, 0);
    Fread (buffer, sizeof (char), we->invf_len, qd->File_invf);

    qd->num_of_terms ++;
    return 0;
  }

  inline void initTerm(WordEntry *_we)
  {
    Abort       = 0;
    CurrDocNum  = 0;  // Note: Document numbers start at 1
    dn_blk      = 0;
    len_blk     = 0;
    skips_taken = 0;
    ptrs_dec    = 0;
    total_added = 0.0;
    hits        = 0;
    we          = _we;
  }

  inline void closeSearch()
  { if (sk)
      fclose (sk);
  }

  /**
   * GRB: Just write a single entry detailing the current position to the
   *      skip dump file
   */
  inline void writeSkipLine(WordEntry *we)
  {
    if (sk)
      {
	fprintf (sk, "%3d : %8ld %6ld \"%.*s\"%*s", term, we->count,
		 we->doc_count, Terms->entry(term)->wordLength(), Terms->entry(term)->word(),
		   indent - Terms->entry(term)->wordLength(), "");
      }
  }

  inline void initSearch()
  {
    num_docs = qd->sd->sdh.num_of_docs;
    if (rqi->MaxTerms == -1)
      num_terms = Terms->size();
    else
      num_terms = rqi->MaxTerms < Terms->size() ? rqi->MaxTerms : Terms->size();
    MoreAccum = 1; // NB: this line isn't used in CosineDecode/GetDocsOp
  }

  inline void writeSkipHeader()
  { 
    if (!skip_header)
      {
	skip_header = 1;
	fprintf (sk, "%s", skip_header_text[qd->id->ifh.skip_mode]);
	fprintf (sk, "\nSkipping method %ld\n", qd->id->ifh.skip_mode);
	switch (qd->id->ifh.skip_mode)
	  {
          case 1:
            fprintf (sk, "Skipping is every %ld docnums\n",
		     qd->id->ifh.params[0]);
	    break;
	  case 2:
	    fprintf (sk, "Max nodes = %ld\nNo skips smaller or equal to %ld\n",
	             qd->id->ifh.params[0], qd->id->ifh.params[1]);
	    break;
	  }
      }
    fprintf (sk, "\n\t\t\t-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n\n");
    for (term = 0; term < Terms->size(); term++)
      if (Terms->entry(term)->wordLength() > indent)
        indent = Terms->entry(term)->wordLength();
  }

  void init(query_data *query, TermList *terms, RankedQueryInfo *qi)
  {
    /* -- note generic call parameters for later use
    */
    Terms = terms;
    qd    = query;
    rqi   = qi;

    /* -- initialise some cosineSearch variables */
    indent = 0;
    Anding = 0;

    /* -- If we've been given a skip dump file to write to, then open
          it
    */
    if (rqi != NULL &&
        rqi->skip_dump != NULL && 
	rqi->skip_dump[0] != '\0')
      {
	char buf[256];
	sprintf (buf, rqi->skip_dump, getpid());
	sk = fopen (buf, "a");
      }
    else
      {
	sk = NULL;
      }

    /* -- if we've got an open skip dump file, then write headers
          to it now
    */
    if (sk != NULL)
      writeSkipHeader();      
  }
};


/* =========================================================================
 * Function: GetDocsOp
 * Description: 
 *      if op == op_term then 
 *        look up term in invf and decode all its entries/documents
 *      if op == op_and_terms then 
 *        go thru L, testing if it exists in invf using skips  
 *        if it does exist then overwrite the entry nearer the start of the list
 *      if op == op_and_or_terms then
 *        go thru L, testing if it exists in invf using skips  
 *        BUT only mark ones which are found - keep list intact
 *      if op == op_diff_terms then     >>>> NOT tested <<<<
 * Input: 
 * Output: 
 * ========================================================================= 
 *
 * GRB: This operates on one term only (pointed to by we); hence Term 
 *      initialisation operations only are performed
 */

DocList *
GetDocsOp (query_data * qd, WordEntry * we, Op_types op, DocList * L)
{
  cosineSearch cs;

  /* GRB: cheats to bluff in cosineSearch; it isn't initialised 
          properly - so don't use functions too glibly!
   */

  DocEntry *src = NULL, *dest = NULL;
  int have_skipping;
  float Weight;

  cs.init(qd, NULL, NULL);    // GRB: in effect this just sets the skip dump file to null;
  cs.initTerm(we);

  /* ... AND empty */
  if (op == op_and_terms && DocList_isEmpty(L))
    {
      if (L)
	Xfree (L);
      return DocList_create (0);
    }

  if (op == op_term)
    L = DocList_create (cs.we->doc_count);

  if (cs.startTerm() != 0)
    return DocList_create (0);

  /* Read from the buffer */
  DECODE_START (cs.buffer, cs.we->invf_len);

  Weight = qd->sd->sdh.num_of_words / (100 * log2 ((float) cs.we->count + 1));

  have_skipping = (cs.skip_size != 0);

  if (op != op_term)
    dest = src = L->member(0);

  cs.prepareSkip();

  /* Go thru the elements of an inverted list for a term */
  /* May not go sequentially if skipping is done */
  while (cs.i < cs.doc_freq && ((op == op_term) || DocList_memberValid(L, src)))
    {
      cs.Anding = (op != op_term); // GRB: I doubt that the op value varies!!!

      (void) cs.skip(&buffer, (src != NULL ? src->docNum() : -1), 1);

      cs.postSkip(&buffer);

      switch (op)
	{
	case op_term:
	  /* --- Add the doc from invf onto the c-set list --- */
	  L->addMember(cs.CurrDocNum, Weight);
	break;

	case op_and_terms:
	case op_and_or_terms:
	  /* --- see if have a match and save it --- */

	  while (DocList_memberValid(L, src) && cs.CurrDocNum > src->docNum())
	    { 
	      src = DocList_nextMember(L, src);
	    }
	  /* Unless at end of list, assert(cs.CurrDocNum <= src->docNum()) */

	  /* --- Accept the doc as a candidate --- */
	  if (DocList_memberValid(L, src) && cs.CurrDocNum == src->docNum())
	    {
	      if (op == op_and_terms)
		{
		  dest->copy(src); // *dest = *src;
		  dest = DocList_nextMember(L, dest);
		}
	      else
		{
		  DocEntry_setOrIncluded(src, 1);
		}
	      src = DocList_nextMember(L, src);	 /* Move onto next one */
	    }

	  /* If the next one to try has already been tried then move on again */
	  if (op == op_and_or_terms)
	    {
	      while (DocList_memberValid(L, src) && DocEntry_getOrIncluded(src))
		src = DocList_nextMember(L, src);
	    }

	  break;

	case op_diff_terms:
	  while (DocList_memberValid(L, src) && src->docNum() < cs.CurrDocNum)
	    {
	      dest->copy(src); // *dest = *src;
	      dest = DocList_nextMember(L, dest);
	      src  = DocList_nextMember(L, src);
	    }
	  if (DocList_memberValid(L, src) && src->docNum() == cs.CurrDocNum)
	    src = DocList_nextMember(L, src);
	  break;

	default:
	  break;
	}			/*switch */

    }				/*while traversing invf */

  if (op == op_and_terms || op == op_diff_terms)
    {
      if (op == op_diff_terms)
	{
	  while (DocList_memberValid(L, src))
	    {	    
	      dest->copy(src); // *dest = *src;
	      dest = DocList_nextMember(L, dest);
	      src  = DocList_nextMember(L, src);
	    }
	  }

      /* update size of c-set */
      DocList_setsize(L, dest - L->member(0));
    }
  DECODE_DONE

    Xfree (cs.buffer);
  ChangeMemInUse (qd, -cs.we->invf_len);

  return (L);
}

class AccumMethod
{
public:
  int doShortCuts;
  virtual int  initMethod(cosineSearch *cs, void *privatedata) = 0;
  virtual void initTerm() = 0;
  virtual void startTerm() = 0;
  virtual int  executeTerm() = 0;
  virtual void endTerm() = 0;
  virtual void endMethod() = 0;
  virtual int  carryOn() = 0;
  virtual int  endTermValue() = 0;
  virtual int  nextDoc() = 0;
  virtual void *endValue() = 0;

  void *cosineDecode(query_data *qd, TermList *Terms, RankedQueryInfo *rqi, void *privatedata);
};

void *AccumMethod::cosineDecode(query_data * qd, TermList * Terms,
		                RankedQueryInfo * rqi, void *privatedata)
{
  cosineSearch cs;

  cs.init(qd, Terms, rqi);
  if (this->initMethod(&cs, privatedata) > 0)
    return NULL;

  cs.initSearch();

  /* For each word in the list . . . */
  for (cs.term = 0; cs.term < cs.num_terms; cs.term++)
    {
      cs.initTerm(&Terms->TE[cs.term].WE);
      this->initTerm();
      cs.writeSkipLine(cs.we);

      COSINE_STARTTERM
      this->startTerm();

      cs.prepareSkip();
      while (cs.i < cs.doc_freq && this->carryOn())
	{
	  cs.skip(&buffer, this->nextDoc(), this->doShortCuts);
	  cs.postSkip(&buffer);

	  cs.Wdt = cs.Count * cs.WordLog;

	  switch (this->executeTerm())
	    {
            case 1:
              goto FastAbort;
	      break;

	    case 2:
	      return NULL;
	      break;
	    }
	}

    FastAbort:
      DECODE_DONE

      Xfree (cs.buffer);
      ChangeMemInUse (qd, -cs.we->invf_len);
      if (cs.sk)
	{
	  if (qd->id->ifh.skip_mode != 0)
	    fprintf (cs.sk, " %6d", cs.skips_taken);
	  fprintf (cs.sk, " %6d %6d %6d%12.2f\n",
		   cs.ptrs_dec, cs.hits, this->endTermValue(), cs.total_added);
	}
      if (cs.Abort)
	break;
      this->endTerm();
    }

  cs.closeSearch();
  this->endMethod();

  return this->endValue();
}

float *
CosineDecode (query_data * qd, TermList * Terms, RankedQueryInfo * rqi)
{
  cosineSearch cs;

  float *AccumulatedWeights;
  int seen_must_match = 0; /* [RJM 07/97: Ranked Required Terms] says whether to 'and' queries or not */

  if (rqi->skip_dump != NULL)
    rqi->skip_dump[0] = '\0';
  cs.init(qd, Terms, rqi);

  cs.initSearch();

  /* Create the array for the documents */
  if (!(AccumulatedWeights = (float *) Xmalloc (cs.num_docs * sizeof (float))))
      return (NULL);
  bzero ((char *) AccumulatedWeights, sizeof (float) * cs.num_docs);
  ChangeMemInUse (qd, sizeof (float) * cs.num_docs);

  /* For each word in the list . . . */
  for (cs.term = 0; cs.term < cs.num_terms; cs.term++)
    {
      unsigned long LastDocNum = 0; /* [RJM 07/97: Ranked Required Terms] */
      unsigned long TempDocNum = 0; /* [RJM 07/97: Ranked Required Terms] */
      unsigned long next_mjr_dn = 0;
      int require_match = Terms->TE[cs.term].require_match; /* [RJM 07/97: Ranked Required Terms] */

      cs.initTerm(&Terms->TE[cs.term].WE);

      COSINE_STARTTERM

      cs.block_counter = 0;
      for (cs.i = 0; cs.i < cs.doc_freq; cs.i++, cs.block_counter++)
	{
	  if (cs.block_counter == cs.skip_size)
	    {
	      cs.block_counter = 0;
	    }
	  if (cs.skip_size && cs.block_counter == 0)
	    if (cs.i + cs.skip_size < cs.doc_freq)
	      {
		int temp;
		next_mjr_dn = buffer.bblock_decode(cs.dn_blk, NULL);
		next_mjr_dn += cs.CurrDocNum;
		temp = buffer.bblock_decode(cs.len_blk, NULL);
	      }
	  if (cs.skip_size && cs.block_counter == cs.skip_size - 1 && cs.i != cs.doc_freq - 1)
	    cs.CurrDocNum = next_mjr_dn;
	  else
	    {
	      cs.diff = buffer.bblock_decode(cs.blk, NULL);
	      cs.CurrDocNum += cs.diff;
	    }

	  if (qd->id->ifh.InvfLevel >= 2)
	    cs.Count = buffer.gamma_decode (NULL);
	  else
	    cs.Count = (int) ((double) cs.we->count / cs.doc_freq);
	  cs.Wdt = cs.Count * cs.WordLog;

	  /* [RJM 07/97: Ranked Required Terms] */
	  /* zero the weights of intermediate documents if this term is required to match */
	  if (require_match == 1) {
	    for (TempDocNum = LastDocNum+1; TempDocNum < cs.CurrDocNum; TempDocNum++)
	      AccumulatedWeights[TempDocNum - 1] = 0.0;
	  }

	  /* [RJM 07/97: Ranked Required Terms] */
	  /* if a term has been required to match before, the weight we are */
	  /* adding to must be non-zero */
	  if (!seen_must_match || (AccumulatedWeights[cs.CurrDocNum - 1] > 0.0)) {
	      AccumulatedWeights[cs.CurrDocNum - 1] += cs.Wqt * cs.Wdt;
	  }

	  qd->num_of_ptrs++;
	  LastDocNum = cs.CurrDocNum; /* [RJM 07/97: Ranked Required Terms] */
	}

      DECODE_DONE

	Xfree (cs.buffer);
      ChangeMemInUse (qd, -cs.we->invf_len);

      /* [RJM 07/97: Ranked Required Terms] */
      /* zero the remaining weights if this term is required to match */
      if (require_match == 1) {
	for (TempDocNum = LastDocNum+1; TempDocNum <= cs.num_docs; TempDocNum++)
	  AccumulatedWeights[TempDocNum - 1] = 0.0;
	seen_must_match = 1;
      }
    }
  return AccumulatedWeights;
}



static int 
doc_comp (void *a, void *b)
{
  return ((Invf_Doc_Entry *) a)->DocNum - ((Invf_Doc_Entry *) b)->DocNum;
}

inline Invf_Doc_Entry *doc_fetch(int docNum, Splay_Tree *sptree)
{ Invf_Doc_Entry  entry;
  Invf_Doc_Entry *reply;

  entry.DocNum = docNum;
  reply = (Invf_Doc_Entry *) SP_member (&entry, sptree);
  return reply;
}

static Invf_Doc_Entry *
get_ide (query_data * qd, Invf_Doc_Entry_Pool * pool)
{
  struct pool *p = (struct pool *) pool->pool;
  if (!p || p->pos == p->length)
    {
      register unsigned size = sizeof (struct pool) +
      (pool->pool_chunks - 1) * sizeof (Invf_Doc_Entry);
      p = (struct pool *) malloc (size);
      if (!p)
	return (Invf_Doc_Entry *) NULL;
      ChangeMemInUse (qd, size);
      p->length = pool->pool_chunks;
      p->pos = 0;
      p->next = (struct pool *) pool->pool;
      pool->pool = p;
    }
  return &(p->ide[p->pos++]);
}


void 
free_ide_pool (query_data * qd, Invf_Doc_Entry_Pool * pool)
{
  while (pool->pool)
    {
      struct pool *p = (struct pool *) pool->pool;
      pool->pool = p->next;
      ChangeMemInUse (qd, -(sizeof (struct pool) + (p->length - 1) *
			    sizeof (Invf_Doc_Entry)));
      free (p);
    }
  pool->pool = NULL;
}

FILE *
open_skip (char *sd)
{
  char buf[256];
  if (sd == NULL || *sd == '\0')
    return NULL;
  sprintf (buf, sd, getpid ());
  return fopen (buf, "a");
}

class SplayMethod : public AccumMethod
{ 
  // Search values
  cosineSearch *cs;
  Splay_Tree *Doc_Set;
  Invf_Doc_Entry_Pool *pool;

  // Term values
  Invf_Doc_Entry *next;

  public:
  SplayMethod();
  int  initMethod(cosineSearch *cs, void *privatedata);
  void initTerm();
  void startTerm();
  int  executeTerm();
  void endTerm();
  void endMethod();
  int  carryOn();
  int  endTermValue();
  int  nextDoc();
  void *endValue();
};

SplayMethod::SplayMethod()
{
  doShortCuts = 0;
}

int SplayMethod::initMethod(cosineSearch *_cs, void *privatedata)
{
  this->cs          = _cs;
  this->Doc_Set     = SP_createset(doc_comp);
  pool              = (Invf_Doc_Entry_Pool *) privatedata;
  pool->pool        = NULL;
  pool->pool_chunks = 1024;
  return 0;
}

void SplayMethod::initTerm()
{
  next = NULL;
}

void SplayMethod::startTerm()
{
  if (this->cs->rqi->MaxAccums != -1 && 
      this->Doc_Set->no_of_items >= this->cs->rqi->MaxAccums)
  {
    cs->MoreAccum = 0;
    cs->Anding = 1;
    this->next = (Invf_Doc_Entry *) SP_get_first (this->Doc_Set);
  }
}

int SplayMethod::executeTerm()
{
  if (cs->Anding)
    {	
      while (next && next->DocNum < cs->CurrDocNum - 1)	
	next = (Invf_Doc_Entry *) SP_get_next (Doc_Set);
      if (next && next->DocNum == cs->CurrDocNum - 1)
	{
	  cs->noteHit(&next->Weight);
	}
      if (next && next->DocNum <= cs->CurrDocNum - 1)
	next = (Invf_Doc_Entry *) SP_get_next (Doc_Set);
      goto NextDocNum;
    }
  if ((cs->mem = doc_fetch(cs->CurrDocNum - 1, Doc_Set)) == NULL)
    {
      cs->hits++;
      if (cs->rqi->MaxAccums != -1 &&
	  Doc_Set->no_of_items >=cs-> rqi->MaxAccums &&
	  !cs->MoreAccum)
	goto NextDocNum;
      if ((cs->mem = get_ide (cs->qd, pool)) == NULL)
	return 2;
      ChangeMemInUse (cs->qd, sizeof (*cs->mem));
      cs->mem->DocNum = cs->CurrDocNum - 1;
      cs->mem->Weight = 0;

      ChangeMemInUse (cs->qd, -Doc_Set->mem_in_use);
      if (SP_insert (cs->mem, Doc_Set) == NULL)
	{
	  Message ("Unable to allocate memory for a splay node");
	  cs->Abort = 1;
	  return 1; // abort return
	}
      ChangeMemInUse (cs->qd, Doc_Set->mem_in_use);
      if (cs->rqi->MaxAccums != -1 &&
	  Doc_Set->no_of_items >= cs->rqi->MaxAccums &&
	  cs->rqi->StopAtMaxAccum)
	cs->Abort = 1;
    }

  cs->mem->Weight += cs->Wqt * cs->Wdt;
  cs->total_added += cs->Wqt * cs->Wdt;

 NextDocNum:
  return 0;    
}

int SplayMethod::carryOn()
{ return (!cs->Anding || next);
}

void SplayMethod::endTerm()
{
  // do nothing
}

int SplayMethod::endTermValue()
{
  return Doc_Set->no_of_items;
}

void SplayMethod::endMethod()
{
  // do nothing
}

int SplayMethod::nextDoc()
{
  return (next != NULL ? next->DocNum : -1);
}

void * SplayMethod::endValue()
{ return (void *) Doc_Set;
}

Splay_Tree *
CosineDecodeSplay (query_data * qd, TermList * Terms,
		   RankedQueryInfo * rqi, Invf_Doc_Entry_Pool * pool)
{
  SplayMethod  sm;

  return (Splay_Tree *) sm.cosineDecode(qd, Terms, rqi, pool);
}


class HashMethod : public AccumMethod
{ 
  // Search values
  cosineSearch *cs;
  Hash_Table *HT;
  int Sorted;

  // Term values
  Invf_Doc_EntryH *next;

  public:
  HashMethod();
  int  initMethod(cosineSearch *cs, void *privatedata);
  void initTerm();
  void startTerm();
  int  executeTerm();
  void endTerm();
  void endMethod();
  int  carryOn();
  int  endTermValue();
  int  nextDoc();
  void *endValue();
};

HashMethod::HashMethod()
{
  doShortCuts = SHORT_CUT;
}

int HashMethod::initMethod(cosineSearch *_cs, void *privatedata)
{
  this->cs          = _cs;
  /* Create the array for the documents */
  Sorted = 0;
  HT = HT_create (cs->qd, cs->rqi->MaxAccums, cs->rqi->HashTblSize);
  if (!HT)
    return 1;
  return 0;  
}

void HashMethod::initTerm()
{
  next = NULL;
}

void HashMethod::startTerm()
{
  if (cs->rqi->MaxAccums != -1 && HT->num >= cs->rqi->MaxAccums)
    {
      cs->MoreAccum = 0;
      cs->Anding = 1;
      if (!Sorted)
	{
	  HT_sort (cs->qd, HT);
	  Sorted = 1;
	}
      next = HT->Head;
    }
}

int HashMethod::executeTerm()
{
  if (cs->Anding)
    {
      while (next && next->DocNum < cs->CurrDocNum - 1)
	next = next->next;
      if (next && next->DocNum == cs->CurrDocNum - 1)
	{
	  cs->noteHit(&next->Weight);
	}
      if (next && next->DocNum <= cs->CurrDocNum - 1)
	next = next->next;
      goto NextDocNum;
    }
  if ((cs->mem = HT_find (HT, cs->CurrDocNum - 1)) == NULL)
    {
      if (cs->rqi->MaxAccums != -1 && HT->num >= cs->rqi->MaxAccums &&
	  !cs->MoreAccum)
	goto NextDocNum;
      if (!HT_insert (HT, cs->CurrDocNum - 1, cs->Wqt * cs->Wdt))
	{
	  if (!HT->Suplimentary_Entries)
	    HT_Sup_create (cs->qd, HT, cs->we->doc_count - cs->i + 1);
	  HT_Sup_Add (HT, cs->CurrDocNum - 1, cs->Wqt * cs->Wdt);
	}
      if (cs->rqi->MaxAccums != -1 && HT->num >= cs->rqi->MaxAccums &&
	  cs->rqi->StopAtMaxAccum)
	cs->Abort = 1;
    }
  else
    {
      cs->noteHit(&cs->mem->Weight);
    }
  
 NextDocNum:
  return 0;
}

int HashMethod::carryOn()
{ return (!cs->Anding || next);
}

void HashMethod::endTerm()
{
  // do nothing
}

int HashMethod::endTermValue()
{
  return HT->num + HT->Suplimentary_Num;
}

void HashMethod::endMethod()
{
  if (!HT->Head)
    HT_sort (cs->qd, HT);
}

int HashMethod::nextDoc()
{
  return (next != NULL ? next->DocNum : -1);
}

void *HashMethod::endValue()
{
  return (void *) HT;
}


Hash_Table *
CosineDecodeHash (query_data * qd, TermList * Terms,
		  RankedQueryInfo * rqi)
{
  HashMethod   hm;
  
  return (Hash_Table *) hm.cosineDecode(qd, Terms, rqi, NULL);
}


class ListMethod : public AccumMethod
{ 
  // Search values
  cosineSearch *cs;
  List_Table *LT;
  int Sorted;
  int has_grown; /* [RJM 07/97: Ranked Required Match] whether elements have been added lately */

  // Term values
  int this_item; /* [RJM 07/97: Ranked Required Match] points to the next possible LT to delete */
  unsigned long LastDocNum;	/* [RJM 07/97: Ranked Required Match] */
  int require_match; /* [RJM 07/97: Ranked Required Match] */
  Invf_Doc_Entry *next;
  int n;

  public:
  ListMethod();
  int  initMethod(cosineSearch *cs, void *privatedata);
  void initTerm();
  void startTerm();
  int  executeTerm();
  void endTerm();
  void endMethod();
  int  carryOn();
  int  endTermValue();
  int  nextDoc();
  void *endValue();
};

ListMethod::ListMethod()
{
  // do nothing
  doShortCuts = 0;
}

int ListMethod::initMethod(cosineSearch *_cs, void *privatedata)
{
  this->cs  = _cs;
  Sorted    = 0;
  has_grown = 0;

  /* Create the list for the documents */
  LT = LT_create (cs->qd, cs->rqi->MaxAccums);
  if (!LT)
    return 1;

  return 0;  
}

void ListMethod::initTerm()
{
  next          = NULL;
  n             = 0;
  LastDocNum    = 0;
  this_item     = 0;
  require_match = cs->Terms->TE[cs->term].require_match;
}

void ListMethod::startTerm()
{
  if (cs->rqi->MaxAccums != -1 && LT->num >= cs->rqi->MaxAccums)
    cs->MoreAccum = 0;

  if ((cs->rqi->MaxAccums != -1 && LT->num >= cs->rqi->MaxAccums) || 
      cs->Anding)
    {
      cs->Anding = 1;
      if (!Sorted || has_grown)
	{
	  LT_sort (LT);
	  LT_pack (LT);
	  Sorted = 1;
	  has_grown = 0;
	}
      next = LT->IDE;
      n = 0;
    }

  /* [RJM 07/97: Ranked Required Terms] */
  /* make sure that the document list is set up correctly for */
  /* requiring a term to match */
  if (require_match == 1) {
    if (!Sorted || has_grown) {
      LT_sort (LT);
      LT_pack (LT);
      Sorted = 1;
      has_grown = 0;
    }
    /* this_item is already set to zero */
  }
}

int ListMethod::executeTerm()
{
  /* [RJM 07/97: Ranked Required Terms] */
  /* if this is a required match we must remove all accumulators */
  /* of intermediate documents */
  if (require_match == 1) {
    while (this_item < LT->num && LT->IDE[this_item].DocNum < cs->CurrDocNum - 1) {
      if (LT->IDE[this_item].DocNum > LastDocNum - 1)
	LT->IDE[this_item].Weight = -1000.0; /* mark for deletion */
      this_item++;
    }
  }
  
  if (cs->Anding)
    {
      while (n < LT->num && next->DocNum < cs->CurrDocNum - 1)
	{
	  next++;
	  n++;
	}
      if (n < LT->num && next->DocNum == cs->CurrDocNum - 1)
	{
	  cs->noteHit(&next->Weight);
	}
      if (n < LT->num && next->DocNum <= cs->CurrDocNum - 1)
	{
	  next++;
	  n++;
	}
      goto NextDocNum;
    }
  if (!cs->Anding)
    {
      if (cs->rqi->MaxAccums != -1 && LT->num >= cs->rqi->MaxAccums &&
	  !cs->MoreAccum)
	goto NextDocNum;
      LT = LT_add (cs->qd, LT, cs->CurrDocNum - 1, cs->Wqt * cs->Wdt);
      if (!LT)
	{
	  cs->Abort = 1;
	  return 1;
	}

      /* the list of accumulators just got bigger */
      has_grown = 1;

      if (cs->rqi->MaxAccums != -1 && LT->num >= cs->rqi->MaxAccums &&
	  cs->rqi->StopAtMaxAccum)
	cs->Abort = 1;
    }
  else if ((cs->mem = LT_find (LT, cs->CurrDocNum - 1)))
    {
      cs->noteHit(&cs->mem->Weight);
    }

 NextDocNum:
  LastDocNum = cs->CurrDocNum;
  return 0;
}

int ListMethod::carryOn()
{ return (!cs->Anding || n < LT->num);
}

void ListMethod::endTerm()
{
  /* [RJM 07/97: Ranked Required Terms] */
  /* remove the remaining accumulators if this term is required to match */
  if (require_match == 1) {
    while (this_item < LT->num) {
      if (LT->IDE[this_item].DocNum > LastDocNum - 1)
	LT->IDE[this_item].Weight = -1000.0; /* mark for deletion */
      this_item++;
    }
    cs->Anding = 1; /* no more documents should be added */

    /* Get rid of any documents which have been deleted */
    LT_pack (LT);
  }
}

int ListMethod::endTermValue()
{
  return LT->num;
}

void ListMethod::endMethod()
{
  // do nothing
}

int ListMethod::nextDoc()
{
  return (next != NULL ? next->DocNum : -1);
}

void *ListMethod::endValue()
{
  return (void *) LT;
}


List_Table *
CosineDecodeList (query_data * qd, TermList * Terms,
		  RankedQueryInfo * rqi)
{
  ListMethod   lm;

  return (List_Table *) lm.cosineDecode(qd, Terms, rqi, NULL);
}
